import numpy as np
import numpy.typing as npt
import scipy.stats
from collections import namedtuple

from nlpstats.correlations.correlations import correlate

FisherResult = namedtuple("FisherResult", ["lower", "upper"])


def fisher(
    X: npt.ArrayLike,
    Z: npt.ArrayLike,
    level: str,
    coefficient: str,
    confidence_level: float = 0.95,
) -> FisherResult:
    _fisher_iv(confidence_level)

    r = correlate(X, Z, level, coefficient)

    # See Bonett and Wright (200) for details
    if coefficient == "pearson":
        b, c = 3, 1
    elif coefficient == "spearman":
        b, c = 3, np.sqrt(1 + r**2 / 2)
    elif coefficient == "kendall":
        b, c = 4, np.sqrt(0.437)
    else:
        raise ValueError(f"Unknown correlation coefficient: {coefficient}")

    if level == "system":
        # The number of systems
        n = X.shape[0]
    elif level == "input":
        # Assume n is the summary-correlation with the largest n.
        # We find that by counting how many non-nans are in each column,
        # then taking the max
        n = (~np.isnan(X)).sum(axis=0).max()
    elif level == "global":
        # The number of non-NaN entries
        n = (~np.isnan(X)).sum()
    else:
        raise Exception(f"Unknown correlation level: {level}")

    alpha = 1 - confidence_level
    if n > b:
        z_r = np.arctanh(r)
        z = scipy.stats.norm.ppf(1.0 - alpha / 2)
        z_l = z_r - z * c / np.sqrt(n - b)
        z_u = z_r + z * c / np.sqrt(n - b)
        r_l = np.tanh(z_l)
        r_u = np.tanh(z_u)
    else:
        r_l, r_u = np.nan, np.nan
    return FisherResult(r_l, r_u)


def _fisher_iv(confidence_level: float) -> None:
    if confidence_level <= 0 or confidence_level >= 1:
        raise ValueError(f"`confidence_level` must be between 0 and 1 (exclusive)")



import numpy as np
import numpy.typing as npt
from collections import namedtuple
from typing import Callable, Union

from nlpstats.correlations.correlations import correlate
from nlpstats.correlations.resampling import resample

BootstrapResult = namedtuple("BootstrapResult", ["lower", "upper", "samples"])


def bootstrap(
    X: npt.ArrayLike,
    Z: npt.ArrayLike,
    level: str,
    coefficient: Union[Callable, str],
    resampling_method: Union[Callable, str],
    paired_inputs: bool = True,
    confidence_level: float = 0.95,
    n_resamples: int = 9999,
) -> BootstrapResult:
    X, Z = _bootstrap_iv(X, Z, level, paired_inputs, confidence_level, n_resamples)

    samples = []
    for _ in range(n_resamples):
        X_s, Z_s = resample((X, Z), resampling_method, paired_inputs=paired_inputs)
        r = correlate(X_s, Z_s, level, coefficient)
        if not np.isnan(r):
            samples.append(r)

    alpha = (1 - confidence_level) / 2
    lower = np.percentile(samples, alpha * 100)
    upper = np.percentile(samples, (1 - alpha) * 100)
    return BootstrapResult(lower, upper, samples)


def _bootstrap_iv(
    X: npt.ArrayLike,
    Z: npt.ArrayLike,
    level: str,
    paired_inputs: bool,
    confidence_level: float,
    n_resamples: int,
):
    X = np.asarray(X)
    Z = np.asarray(Z)

    if not paired_inputs and level in {"input", "global"}:
        raise ValueError(
            f"`paired_inputs` must be `True` for input- or global-level correlations"
        )

    if confidence_level <= 0 or confidence_level >= 1:
        raise ValueError(f"`confidence_level` must be between 0 and 1 (exclusive)")

    if n_resamples <= 0:
        raise ValueError(f"`n_resamples` must be a positive integer")

    return X, Z



from nlpstats.correlations.fisher import fisher
from nlpstats.correlations.williams import williams_test
from nlpstats.correlations.resampling import permute, resample
from nlpstats.correlations.correlations import (
    correlate,
    global_level,
    input_level,
    system_level,
)
from nlpstats.correlations.bootstrap import bootstrap
from nlpstats.correlations.permutation import permutation_test



import numpy as np
import numpy.typing as npt
from typing import Callable, List, Sequence, Tuple, Union


def _resample_systems(matrices: List[np.ndarray], **kwargs) -> List[np.ndarray]:
    _resample_systems_iv(matrices)
    m = matrices[0].shape[0]
    rows = np.random.choice(m, m, replace=True)
    return [matrix[rows] for matrix in matrices]


def _resample_systems_iv(matrices: List[np.ndarray]) -> None:
    m = matrices[0].shape[0]
    for matrix in matrices:
        if matrix.shape[0] != m:
            raise ValueError("Input `matrices` all must have the same number of rows")


def _resample_inputs(
    matrices: List[np.ndarray], paired_inputs: bool
) -> List[np.ndarray]:
    _resample_inputs_iv(matrices, paired_inputs)
    if paired_inputs:
        n = matrices[0].shape[1]
        cols = np.random.choice(n, n, replace=True)
        return [matrix[:, cols] for matrix in matrices]
    else:
        resamples = []
        for matrix in matrices:
            n = matrix.shape[1]
            cols = np.random.choice(n, n, replace=True)
            resamples.append(matrix[:, cols])
        return resamples


def _resample_inputs_iv(matrices: List[np.ndarray], paired_inputs: bool) -> None:
    if paired_inputs:
        n = matrices[0].shape[1]
        for matrix in matrices:
            if matrix.shape[1] != n:
                raise ValueError(
                    "Input `matrices` all must have the same number of columns"
                )


def _resample_both(matrices: List[np.ndarray], paired_inputs: bool) -> List[np.ndarray]:
    _resample_both_iv(matrices, paired_inputs)
    matrices = _resample_systems(matrices)
    return _resample_inputs(matrices, paired_inputs)


def _resample_both_iv(matrices: List[np.ndarray], paired_inputs: bool) -> None:
    _resample_systems_iv(matrices)
    _resample_inputs_iv(matrices, paired_inputs)


def resample(
    matrices: Union[npt.ArrayLike, Sequence[npt.ArrayLike]],
    resampling_method: Union[Callable, str],
    paired_inputs: bool = True,
) -> Union[np.ndarray, Tuple[np.ndarray, ...]]:
    matrices, resampling_method = _resample_iv(matrices, resampling_method)
    resamples = resampling_method(matrices, paired_inputs=paired_inputs)
    if len(resamples) == 1:
        return resamples[0]
    return tuple(resamples)


def _resample_iv(
    matrices: Union[npt.ArrayLike, Sequence[npt.ArrayLike]],
    resampling_method: Union[Callable, str],
) -> Tuple[Sequence[np.ndarray], Callable]:
    # If the input is just one matrix, wrap it into a list
    if isinstance(matrices, np.ndarray):
        matrices = [matrices]

    # Ensure all are numpy arrays
    for matrix in matrices:
        if not isinstance(matrix, np.ndarray):
            raise TypeError(f"Input `matrices` must all be of type `np.ndarray`")

    if isinstance(resampling_method, str):
        if resampling_method == "systems":
            resampling_method = _resample_systems
        elif resampling_method == "inputs":
            resampling_method = _resample_inputs
        elif resampling_method == "both":
            resampling_method = _resample_both
        else:
            raise ValueError(f"Unknown resampling method: {resampling_method}")

    return matrices, resampling_method


def _permutation_iv(X: np.ndarray, Y: np.ndarray):
    if X.ndim != 2:
        raise ValueError(f"`X` must be two-dimensional")
    if Y.ndim != 2:
        raise ValueError(f"`Y` must be two-dimensional")
    if X.shape != Y.shape:
        raise ValueError(f"`X` and `Y` must be the same shape")


def _permute_systems(X: np.ndarray, Y: np.ndarray) -> Tuple[np.ndarray, np.ndarray]:
    _permutation_iv(X, Y)

    # Do not modify the original matrices
    X_p = X.copy()
    Y_p = Y.copy()

    m = X.shape[0]
    mask = (np.random.rand(m, 1) > 0.5).reshape((m,))
    X_p[mask] = Y[mask]
    Y_p[mask] = X[mask]
    return X_p, Y_p


def _permute_inputs(X: np.ndarray, Y: np.ndarray) -> Tuple[np.ndarray, np.ndarray]:
    _permutation_iv(X, Y)

    # Do not modify the original matrices
    X_p = X.copy()
    Y_p = Y.copy()

    n = X.shape[1]
    mask = (np.random.rand(1, n) > 0.5).reshape((n,))
    X_p[:, mask] = Y[:, mask]
    Y_p[:, mask] = X[:, mask]
    return X_p, Y_p


def _permute_both(X: np.ndarray, Y: np.ndarray) -> Tuple[np.ndarray, np.ndarray]:
    X, Y = _permute_systems(X, Y)
    return _permute_inputs(X, Y)


def _permute_iv(
    X: npt.ArrayLike, Y: npt.ArrayLike, permutation_method: str
) -> Tuple[np.ndarray, np.ndarray, Callable]:
    X = np.asarray(X)
    Y = np.asarray(Y)

    if isinstance(permutation_method, str):
        if permutation_method == "systems":
            permutation_method = _permute_systems
        elif permutation_method == "inputs":
            permutation_method = _permute_inputs
        elif permutation_method == "both":
            permutation_method = _permute_both
        else:
            raise ValueError(f"Unknown permutation method: {permutation_method}")

    return X, Y, permutation_method


def permute(
    X: npt.ArrayLike, Y: npt.ArrayLike, permutation_method: str
) -> Tuple[np.ndarray, np.ndarray]:
    X, Y, permutation_method = _permute_iv(X, Y, permutation_method)
    return permutation_method(X, Y)



import numpy as np
import numpy.typing as npt
from collections import namedtuple
from typing import Callable, Tuple, Union

from nlpstats.correlations.correlations import correlate
from nlpstats.correlations.resampling import permute

PermutationResult = namedtuple("PermutationResult", ["pvalue", "samples"])


def _standardize(X: np.ndarray) -> np.ndarray:
    return (X - np.nanmean(X)) / np.nanstd(X)


def permutation_test(
    X: npt.ArrayLike,
    Y: npt.ArrayLike,
    Z: npt.ArrayLike,
    level: str,
    coefficient: Union[Callable, str],
    permutation_method: Union[Callable, str],
    paired_inputs: bool = True,
    alternative: str = "two-sided",
    n_resamples: int = 9999,
) -> PermutationResult:
    X, Y, Z = _permutation_test_iv(
        X, Y, Z, level, paired_inputs, alternative, n_resamples
    )

    # The data needs to be standardized so the metrics are on the same scale.
    X = _standardize(X)
    Y = _standardize(Y)
    Z = _standardize(Z)

    observed = correlate(X, Z, level, coefficient) - correlate(Y, Z, level, coefficient)

    samples = []
    count = 0
    for _ in range(n_resamples):
        X_p, Y_p = permute(X, Y, permutation_method)
        sample = correlate(X_p, Z, level, coefficient) - correlate(
            Y_p, Z, level, coefficient
        )
        samples.append(sample)

        if alternative == "two-sided":
            if abs(sample) >= abs(observed):
                count += 1
        elif alternative == "greater":
            if sample >= observed:
                count += 1
        elif alternative == "less":
            if sample <= observed:
                count += 1
        else:
            raise ValueError(f"Unknown alternative: {alternative}")

    pvalue = count / len(samples)
    return PermutationResult(pvalue, samples)


def _permutation_test_iv(
    X: npt.ArrayLike,
    Y: npt.ArrayLike,
    Z: npt.ArrayLike,
    level: str,
    paired_inputs: bool,
    alternative: str,
    n_resamples: int,
) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
    X = np.asarray(X)
    Y = np.asarray(Y)
    Z = np.asarray(Z)

    if not paired_inputs and level in {"input", "global"}:
        raise ValueError(
            f"`paired_inputs` must be `True` for input- or global-level correlations"
        )

    if alternative not in {"two-sided", "greater", "less"}:
        raise ValueError(f"Unknown alternative: {alternative}")

    if n_resamples <= 0:
        raise ValueError(f"`n_resamples` must be a positive integer")

    return X, Y, Z



import numpy as np
import numpy.typing as npt
from scipy.stats import kendalltau, pearsonr, spearmanr
from typing import Callable, Tuple, Union


def _kendalltau(*args):
    return kendalltau(*args)[0]


def _pearsonr(*args):
    return pearsonr(*args)[0]


def _spearmanr(*args):
    return spearmanr(*args)[0]


def _coefficient_iv(coefficient: Union[Callable, str]) -> Callable:
    if isinstance(coefficient, str):
        if coefficient == "kendall":
            return _kendalltau
        elif coefficient == "pearson":
            return _pearsonr
        elif coefficient == "spearman":
            return _spearmanr
        else:
            raise ValueError(f"Unknown correlation coefficient: {coefficient}")
    return coefficient


def _correlation_iv(
    X: npt.ArrayLike, Z: npt.ArrayLike, coefficient: Union[Callable, str]
) -> Tuple[np.ndarray, np.ndarray, Callable]:
    X = np.asarray(X)
    Z = np.asarray(Z)
    if X.ndim != 2:
        raise ValueError(f"`X` must be two-dimensional")
    if Z.ndim != 2:
        raise ValueError(f"`Z` must be two-dimensional")
    if X.shape[0] != Z.shape[0]:
        raise ValueError(f"`X` and `Z` must have the same number of rows")

    coefficient = _coefficient_iv(coefficient)
    return X, Z, coefficient


def system_level(
    X: npt.ArrayLike, Z: npt.ArrayLike, coefficient: Union[Callable, str]
) -> float:
    X, Z, coefficient = _system_level_iv(X, Z, coefficient)
    x = np.nanmean(X, axis=1)
    z = np.nanmean(Z, axis=1)
    return coefficient(x, z)


def _system_level_iv(
    X: npt.ArrayLike, Z: npt.ArrayLike, coefficient: Union[Callable, str]
) -> Tuple[np.ndarray, np.ndarray, Callable]:
    X, Z, coefficient = _correlation_iv(X, Z, coefficient)

    n1 = X.shape[1]
    n2 = Z.shape[1]
    if (np.isnan(X).sum(axis=1) == n1).any():
        raise ValueError(f"`X` must not have a row of all NaN")
    if (np.isnan(Z).sum(axis=1) == n2).any():
        raise ValueError(f"`Z` must not have a row of all NaN")

    return X, Z, coefficient


def input_level(
    X: npt.ArrayLike,
    Z: npt.ArrayLike,
    coefficient: Union[Callable, str],
) -> float:
    X, Z, coefficient = _input_level_iv(X, Z, coefficient)
    n = X.shape[1]
    rs = []
    for j in range(n):
        x, z = X[:, j], Z[:, j]

        # Remove any NaN. Because the input validation checks to
        # ensure the NaNs are in identical locations, these vectors
        # will be paired
        x = x[~np.isnan(x)]
        z = z[~np.isnan(z)]
        r = coefficient(x, z)
        if not np.isnan(r):
            rs.append(r)

    if len(rs) == 0:
        return np.nan
    return np.mean(rs)


def _input_level_iv(
    X: npt.ArrayLike,
    Z: npt.ArrayLike,
    coefficient: Union[Callable, str],
) -> Tuple[np.ndarray, np.ndarray, Callable]:
    X, Z, coefficient = _correlation_iv(X, Z, coefficient)

    if X.shape[1] != Z.shape[1]:
        raise ValueError(f"`X` and `Z` must have the same number of columns")

    m1 = X.shape[0]
    m2 = Z.shape[0]
    if (np.isnan(X).sum(axis=0) == m1).any():
        raise ValueError(f"`X` must not have a column of all NaN")
    if (np.isnan(Z).sum(axis=0) == m2).any():
        raise ValueError(f"`Z` must not have a column of all NaN")

    if np.not_equal(np.isnan(X), np.isnan(Z)).any():
        raise ValueError(f"`X` and `Z` must have identical NaN locations")

    return X, Z, coefficient


def global_level(
    X: npt.ArrayLike,
    Z: npt.ArrayLike,
    coefficient: Union[Callable, str],
) -> float:
    X, Z, coefficient = _global_level_iv(X, Z, coefficient)

    # Flatten into vectors
    x, z = X.flatten(), Z.flatten()

    # Remove NaNs. x and z will still be paired because
    # the input validation checks to make sure the NaNs are in
    # identical locations
    x = x[~np.isnan(x)]
    z = z[~np.isnan(z)]

    return coefficient(x, z)


def _global_level_iv(
    X: npt.ArrayLike,
    Z: npt.ArrayLike,
    coefficient: Union[Callable, str],
) -> Tuple[np.ndarray, np.ndarray, Callable]:
    X, Z, coefficient = _correlation_iv(X, Z, coefficient)

    if X.shape[1] != Z.shape[1]:
        raise ValueError(f"`X` and `Z` must have the same number of columns")

    if X.size - np.isnan(X).sum() < 2:
        raise ValueError(f"`X` must have at least 2 non-NaN values")
    if Z.size - np.isnan(Z).sum() < 2:
        raise ValueError(f"`Z` must have at least 2 non-NaN values")

    if np.not_equal(np.isnan(X), np.isnan(Z)).any():
        raise ValueError(f"`X` and `Z` must have identical NaN locations")

    return X, Z, coefficient


def correlate(
    X: npt.ArrayLike,
    Z: npt.ArrayLike,
    level: str,
    coefficient: Union[Callable, str],
) -> float:
    X, Z, level = _correlate_iv(X, Z, level)
    return level(X, Z, coefficient)


def _correlate_iv(
    X: npt.ArrayLike, Z: npt.ArrayLike, level: str
) -> Tuple[np.ndarray, np.ndarray, Callable]:
    X = np.asarray(X)
    Z = np.asarray(Z)

    if level == "system":
        level = system_level
    elif level == "input":
        level = input_level
    elif level == "global":
        level = global_level
    else:
        raise ValueError(f"Unknown correlation level: {level}")

    return X, Z, level



import setuptools


# version.py defines the VERSION variable.
# We use exec here so we don't import the whole package whilst setting up.
VERSION = {}
with open("nlpstats/version.py", "r") as version_file:
    exec(version_file.read(), VERSION)


setuptools.setup(
    name="nlpstats",
    version=VERSION["VERSION"],
    author="Daniel Deutsch",
    description="A statistical toolbox for NLP",
    url="https://github.com/danieldeutsch/nlpstats",
    packages=setuptools.find_packages(exclude=["tests", "tests.*"]),
    python_requires=">=3.7",
    include_package_data=True,
    install_requires=[
        "numpy>=1.20.0",
        "scipy",
    ],
)



import numpy as np
import numpy.typing as npt
import scipy.stats
from collections import namedtuple
from typing import Callable, Union

from nlpstats.correlations.correlations import correlate

WilliamsResult = namedtuple("WilliamsResult", ["pvalue"])


def williams_test(
    X: npt.ArrayLike,
    Y: npt.ArrayLike,
    Z: npt.ArrayLike,
    level: str,
    coefficient: Union[Callable, str],
    alternative: str = "two-sided",
) -> WilliamsResult:
    # In the math, Z is metric 1. We take the absolute value of the correlations because
    # it does not matter whether they are positively or negatively correlated with each other. The WMT scripts
    # do the same before calling r.test
    r12 = abs(correlate(X, Z, level, coefficient))
    r13 = abs(correlate(Y, Z, level, coefficient))
    r23 = abs(correlate(X, Y, level, coefficient))

    if level == "system":
        # The number of systems
        n = X.shape[0]
    elif level == "input":
        # Assume n is the summary-correlation with the largest n.
        # We find that by counting how many non-nans are in each column,
        # then taking the max
        n = (~np.isnan(X)).sum(axis=0).max()
    elif level == "global":
        # The number of non-NaN entries
        n = (~np.isnan(X)).sum()
    else:
        raise Exception(f"Unknown correlation level: {level}")

    # Implementation based on https://github.com/cran/psych/blob/master/R/r.test.R
    diff = r12 - r13
    det = 1 - (r12**2) - (r23**2) - (r13**2) + (2 * r12 * r23 * r13)
    av = (r12 + r13) / 2
    cube = (1 - r23) ** 3
    t2 = diff * np.sqrt(
        (n - 1) * (1 + r23) / ((2 * (n - 1) / (n - 3)) * det + av**2 * cube)
    )

    # r.test implicitly assumes that r12 > r13 because it takes the absolute value of the
    # t statistic. Since we don't, we have to have special handling for one-tailed tests
    # so we don't map a negative t statistic to a positive one.
    if alternative == "two-sided":
        pvalue = scipy.stats.t.sf(abs(t2), n - 3) * 2
    elif alternative == "greater":
        pvalue = scipy.stats.t.sf(t2, n - 3)
    elif alternative == "less":
        pvalue = scipy.stats.t.sf(-1 * t2, n - 3)
    else:
        raise ValueError(f"Unknown alternative: {alternative}")

    return WilliamsResult(pvalue)
